"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const util = require('util');
const utils_1 = require("utils");
class Pluggable {
    /**
     *
     * @param {typeof Pluggable} pluggable
     * @private
     */
    static _initialize(pluggable) {
        if (pluggable._initialized)
            return;
        pluggable._initialized = true;
        pluggable._hooks = {
            pre: {},
            post: {},
        };
        pluggable._declares = [];
        pluggable._subscribers = [];
    }
    /**
     * @constructor
     * @public
     */
    constructor() {
        const ThisClass = this.constructor;
        Pluggable._initialize(ThisClass);
        this._hooks = utils_1.copy(ThisClass._hooks);
        ThisClass._subscribers.push(this);
        for (let declares of ThisClass._declares) {
            this[declares.path] = typeof declares.value === 'function' ? declares.value() : utils_1.copy(declares.value);
        }
    }
    /**
     *
     * @param {PluggablePeriod} period
     * @param {string | string[]} name
     * @param {Function} funcs
     * @private
     */
    _addHooks(period, name, ...funcs) {
        const hooks = this._hooks[period];
        if (typeof name === 'string') {
            hooks[name].push(...utils_1.flat(funcs));
        }
        else if (name instanceof Array) {
            for (let n of name) {
                hooks[n].push(...utils_1.flat(funcs));
            }
        }
        else {
            throw ReferenceError('Hook\'s name is string or array of string');
        }
    }
    /**
     *
     * @param {string | string[]} name
     * @param {Function} funcs
     * @public
     */
    pre(name, ...funcs) {
        this._addHooks('pre', name, ...funcs);
    }
    /**
     *
     * @param {string | string[]} name
     * @param {Function} funcs
     * @public
     */
    post(name, ...funcs) {
        this._addHooks('post', name, ...funcs);
    }
    /**
     *
     * @param {PluggablePeriod} period
     * @param {string} name
     * @param {...*}argus
     * @return {Promise<void>}
     * @private
     */
    async _triggerHooks(period, name, ...argus) {
        const plugins = this._hooks[period][name];
        if (plugins) {
            for (let plugin of plugins) {
                await plugin.apply(this, argus);
            }
        }
    }
    /**
     *
     * @param {string | string[]} name
     * @param {Function} funcs
     * @public
     */
    static pre(name, ...funcs) {
        Pluggable.prototype.pre.apply(this, arguments);
    }
    /**
     *
     * @param {string | string[]} name
     * @param {Function} funcs
     */
    static post(name, ...funcs) {
        Pluggable.prototype.post.apply(this, arguments);
    }
    /**
     *
     * @param {PluggablePeriod} period
     * @param {string | string[]} name
     * @param {Function} funcs
     * @private
     */
    static _addHooks(period, name, ...funcs) {
        this.prototype._addHooks.apply(this, arguments);
        for (let subscriber of this._subscribers) {
            this.prototype._addHooks.apply(subscriber, arguments);
        }
    }
    /**
     *
     * @param {string} path
     * @param value
     */
    static declare(path, value) {
        this._declares.push({ path, value });
        for (let subscriber of this._subscribers) {
            subscriber[path] = typeof value === 'function' ? value() : utils_1.copy(value);
        }
    }
    static wrapHook(funcName, hookName, periodsSet) {
        if (this === Pluggable) {
            throw ReferenceError('Pluggable is an interface.');
        }
        Pluggable._initialize(this);
        if (!periodsSet) {
            periodsSet = hookName;
            hookName = funcName;
        }
        const periods = periodsSet.split(' ')
            .map(e => e.trim())
            .filter(e => /pre|post/.test(e));
        if (periods.length !== 0) {
            const hooks = this._hooks;
            periods.forEach((period) => {
                hooks[period][hookName] = [];
            });
            const originalMethod = this.prototype[funcName];
            this.prototype[funcName] = async function (...args) {
                let result = await this._triggerHooks('pre', hookName, ...args)
                    .then(async () => typeof originalMethod === 'function' ? await originalMethod.apply(this, args) : args[0]);
                await this._triggerHooks('post', hookName, result);
                return result;
            };
        }
        else {
            throw SyntaxError(`wrapped methods needs periods.`);
        }
    }
}
exports.default = Pluggable;
//# sourceMappingURL=index.js.map