const util = require('util')
import {copy, flat} from 'utils'

export type PluggablePeriod = 'pre' | 'post'
export type PluggablePeriodSet = PluggablePeriod | 'pre post'
export type PluggableHooks = { pre: { [key: string]: Function[] }, post: { [key: string]: Function[] } }

export default class Pluggable {

  static _initialized: boolean
  static _hooks: PluggableHooks
  static _declares: { path: string, value: any }[]
  static _subscribers: Pluggable[]

  private _hooks: PluggableHooks
  [key: string]: any

  /**
   *
   * @param {typeof Pluggable} pluggable
   * @private
   */
  private static _initialize (pluggable: typeof Pluggable): void {
    if (pluggable._initialized) return
    pluggable._initialized = true
    pluggable._hooks = {
      pre: {},
      post: {},
    }
    pluggable._declares = []
    pluggable._subscribers = []
  }

  /**
   * @constructor
   * @public
   */
  constructor () {
    const ThisClass: typeof Pluggable = <any>this.constructor
    Pluggable._initialize(ThisClass)
    this._hooks = copy(ThisClass._hooks)
    ThisClass._subscribers.push(this)
    for (let declares of ThisClass._declares) {
      this[declares.path] = typeof declares.value === 'function' ? declares.value() : copy(declares.value)
    }
  }

  /**
   *
   * @param {PluggablePeriod} period
   * @param {string | string[]} name
   * @param {Function} funcs
   * @private
   */
  private _addHooks (period: PluggablePeriod, name: string | string[], ...funcs: Function[]) {
    const hooks = this._hooks[period]
    if (typeof name === 'string') {
      hooks[name].push(...flat(funcs))
    }
    else if (name instanceof Array) {
      for (let n of name) {
        hooks[n].push(...flat(funcs))
      }
    } else {
      throw ReferenceError('Hook\'s name is string or array of string')
    }

  }

  /**
   *
   * @param {string | string[]} name
   * @param {Function} funcs
   * @public
   */
  public pre (name: string | string[], ...funcs: Function[]) {
    this._addHooks('pre', name, ...funcs)
  }

  /**
   *
   * @param {string | string[]} name
   * @param {Function} funcs
   * @public
   */
  public post (name: string | string[], ...funcs: Function[]) {
    this._addHooks('post', name, ...funcs)
  }

  /**
   *
   * @param {PluggablePeriod} period
   * @param {string} name
   * @param {...*}argus
   * @return {Promise<void>}
   * @private
   */
  private async _triggerHooks (period: PluggablePeriod, name: string, ...argus: any[]) {
    const plugins = this._hooks[period][name]
    if (plugins) {
      for (let plugin of plugins) {
        await plugin.apply(this, argus)
      }
    }
  }

  /**
   *
   * @param {string | string[]} name
   * @param {Function} funcs
   * @public
   */
  public static pre (name: string | string[], ...funcs: Function[]) {
    Pluggable.prototype.pre.apply(this, arguments)
  }

  /**
   *
   * @param {string | string[]} name
   * @param {Function} funcs
   */
  public static post (name: string | string[], ...funcs: Function[]) {
    Pluggable.prototype.post.apply(this, arguments)
  }

  /**
   *
   * @param {PluggablePeriod} period
   * @param {string | string[]} name
   * @param {Function} funcs
   * @private
   */
  private static _addHooks (period: PluggablePeriod, name: string | string[], ...funcs: Function[]) {
    this.prototype._addHooks.apply(this, arguments)
    for (let subscriber of this._subscribers) {
      this.prototype._addHooks.apply(subscriber, arguments)
    }
  }

  /**
   *
   * @param {string} path
   * @param value
   */
  static declare (path: string, value: any) {
    this._declares.push({path, value})
    for (let subscriber of this._subscribers) {
      subscriber[path] = typeof value === 'function' ? value() : copy(value)
    }
  }

  /**
   *
   * @param {string} funcName
   * @param {string} hookName
   * @param {PluggablePeriodSet} periodsSet
   */
  static wrapHook (funcName: string, hookName: string, periodsSet: PluggablePeriodSet): void;

  /**
   *
   * @param {string} funcAndHookName
   * @param {PluggablePeriodSet} periodsSet
   * @param  _
   */
  static wrapHook (funcAndHookName: string, periodsSet: PluggablePeriodSet, _: never): void;

  static wrapHook (funcName: string, hookName: string | PluggablePeriodSet, periodsSet?: PluggablePeriodSet) {
    if (this === Pluggable) {
      throw ReferenceError('Pluggable is an interface.')
    }
    Pluggable._initialize(this)
    if (!periodsSet) {
      periodsSet = <any>hookName
      hookName = funcName
    }
    const periods = periodsSet.split(' ')
      .map(e => e.trim())
      .filter(e => /pre|post/.test(e))
    if (periods.length !== 0) {
      const hooks = this._hooks
      periods.forEach((period: PluggablePeriod) => {
        hooks[period][hookName] = []
      })
      const originalMethod = this.prototype[funcName]
      this.prototype[funcName] = async function (...args: any[]) {
        let result = await this._triggerHooks('pre', hookName, ...args)
          .then(async () => typeof originalMethod === 'function' ? await originalMethod.apply(this, args) : args[0])
        await this._triggerHooks('post', hookName, result)
        return result
      }
    } else {
      throw SyntaxError(`wrapped methods needs periods.`)
    }
  }

}

export interface DefaultPluggin<T> {
  (param: T): Promise<T>
}
